#!/usr/bin/env python
# -*- coding:utf-8 -*-
import getopt
import json
import sys

from value import *
from config import load_config
from utils import grafana_table_format, grafana_timeserie_format

import web
import restsql as rql

logger = logging.getLogger("main")
logger.setLevel(LOG_LEVEL)

urls = (
    "", "ConnectionView",
    "/", "ConnectionView",
    "/query", "QueryView",
    "/search", "VariableView",
    "/find_options", "FieldView",
    "/find_tables", "TableView"
)


class QueryView(object):
    """
    查询query接口
    request data:
    {
        'format': '...',
        (for grafana.timeserie
        'beginTime': ,
        'endTime':
        )
        'data':
        [
            {
                'query':
                {
                    ...
                },

                (for grafana.timeserie
                'timeShift': ,
                'timeShiftDimension': '',
                'timeAgg': ,
                'timeAggDimension': ''
                )
            },
            {
                ...
            }
        ]
    }
    """
    TABLE = 'grafana.table'
    TIMESERIE = 'grafana.timeserie'
    NORMAL = 'normal'

    def POST(self):
        logger.info("QueryView active.")
        try:
            data = json.loads(web.data())
        except Exception as e:
            logging.exception(e)
            raise web.internalerror('解析json时发生错误')
        grafana_type = data.get('format', 'normal')
        begin_time = data.get('beginTime', None)
        end_time = data.get('endTime', None)
        response_data = []  # 存储多query查询的结果
        for query_data in data.get('data'):
            # 获取相关信息
            query_dict = query_data.get('query')
            time_shift = query_data.get('timeShift', None)
            time_shift_dimension = query_data.get('timeShiftDimension', None)
            time_agg = query_data.get('timeAgg', None)
            time_agg_dimension = query_data.get('timeAggDimension', None)
            # 查询
            try:
                result_df = rql.RestSQL.query(query_dict)
            except rql.RestSqlExceptionBase as e:
                logger.error('query查询失败: %s %s', e.code, e.message)
                raise web.internalerror('query查询失败: \
                                         错误代码: {}; \
                                         错误信息: {}'.format(e.code, e.message))
            except Exception as e:
                logger.error('查询时发生未知异常: %s', e)
                logging.exception(e)
                raise web.internalerror('查询时发生未知异常: {}'.format(e))
            query_result = None
            try:
                if grafana_type == QueryView.NORMAL:
                    query_result = json.loads(result_df.to_json(orient='records'))
                elif grafana_type == QueryView.TABLE:
                    query_result = grafana_table_format(result_df)
                elif grafana_type == QueryView.TIMESERIE:
                    query_result = grafana_timeserie_format(result_df, time_shift, time_shift_dimension, time_agg,
                                                            time_agg_dimension, begin_time, end_time)
                else:
                    raise rql.RunningException(1, 'query查询后格式化遇到未知grafana type: {}', grafana_type)
            except rql.RestSqlExceptionBase as e:
                logger.error(e.message)
                raise web.internalerror('query数据格式化失败: \
                                         错误代码: {}; \
                                         错误信息: {}'.format(e.code, e.message))
            response_data.append(query_result)
        return json.dumps({
            'status': 'ok',
            'data': response_data
        })


class VariableView(object):
    """
    查询variable接口
    request:
    {
        'target': '...' # example: {"select": "id", "from": "user"}
    }
    response:
        ['...', ...]
    """

    def POST(self):
        logger.info("VariableView active.")
        try:
            data = json.loads(web.data())
        except Exception as e:
            logger.error('[VariableView] 解析json时发生错误')
            logging.exception(e)
            raise web.internalerror('[VariableView] 解析json时发生错误，请查看运行日志')
        # 提取变量
        target = data.get('target')
        if target is None:
            raise web.internalerror('[VariableView] 查询格式不正确: 无target字段')
        select_part = target.get('select', None)
        from_part = target.get('from', None)
        if select_part is None:
            raise web.internalerror('[VariableView] 查询格式不正确，需要select字段')
        if from_part is None:
            raise web.internalerror('[VariableView] 查询格式不正确，需要from字段')
        # 查询
        query_dict = self._gen_query(select_part, from_part)
        try:
            result_df = rql.RestSQL.query(query_dict)
        except rql.RestSqlExceptionBase as e:
            logger.error('[VariableView] 查询失败')
            logging.exception(e)
            raise web.internalerror('[VariableView] 查询失败: \
                        错误代码: {}; \
                        错误信息: {}'.format(e.code, e.message))
        except Exception as e:
            logger.error('[VariableView] 查询失败: 未知异常')
            logging.exception(e)
            raise web.internalerror('[VariableView] 查询失败: 未知异常，请查看运行日志')
        result_dict = result_df.to_dict()
        return json.dumps({
            'status': 'ok',
            'data': tuple(result_dict.get(select_part).values())
        })

    def _gen_query(self, select_part, from_part):
        """
        根据variable查询的输入语句，生成restql查询。
        :param select_part: select字段
        :param from_part: from字段
        :return:
        """
        return {
            'select': {
                'from': from_part,
                'fields': [select_part],
                'group_by': [select_part],
                'limit': 10000
            },
            'sort': [select_part],
            'fields': ['{}@{}'.format(select_part, select_part)],
            'limit': 10000
        }


class TableView(object):
    """
    查询可选表的接口
    response:
    {
        'status': 'ok',
        'data': ['...', ...]
    }
    """

    def GET(self):
        logger.info("TableView active.")
        return json.dumps({
            'status': 'ok',
            'data': list(rql.index4table.keys())
        })


class FieldView(object):
    """
    查询表字段接口
    request:
        {
            'tableName': '...'<String>
        }
    response:
        ['...', ...]
    """

    def POST(self):
        logger.info("FieldView active.")
        try:
            data = json.loads(web.data())
        except Exception as e:
            logger.error('[FieldView] 解析json时发生错误')
            logging.exception(e)
            raise web.internalerror('[FieldView] 解析json时发生错误，请查看运行日志')
        table_name = data.get("tableName")
        table = rql.index4table.get(table_name, None)
        if table is None:
            raise web.internalerror('[FieldView] 需查询字段的表{}不存在'.format(table_name))
        return json.dumps({
            'status': 'ok',
            'data': list(table.fields.keys())
        })


class ConnectionView(object):
    """
    用于测试后端运行中
    """

    def GET(self):
        logger.info("ConnectionView active.")
        return 'Connection successful!'


app = web.application(urls, globals())

if __name__ == "__main__":
    logger.info('restsql server为您服务!')
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hc:l:", ["help", "config", "level="])
    except getopt.GetoptError as e:
        logger.critical('Usage: python %s. Error when parsing arguments', sys.argv[0])
        logging.exception(e)
        sys.exit(2)
    path = None
    for opt, arg in opts:
        if opt in ('-h', '--help'):
            print('python {}'.format(sys.argv[0]))
            sys.exit()
        elif opt in ('-c', '--config'):
            path = arg
        elif opt in ('-l', '--level'):
            if arg == 'DEBUG':
                LOG_LEVEL = logging.DEBUG
            elif arg == 'INFO':
                LOG_LEVEL = logging.INFO
            elif arg == 'WARNING':
                LOG_LEVEL = logging.WARNING
            elif arg == 'ERROR':
                LOG_LEVEL = logging.ERROR
            elif arg == 'CRITICAL':
                LOG_LEVEL = logging.CRITICAL
            else:
                logger.critical('Error: level should in [\'DEBUG\', \'INFO\', \'WARNING\', \'ERROR\', \'CRITICAL\']')
                sys.exit(3)
            logger.setLevel(LOG_LEVEL)
        else:
            logger.critical('Error: Unknown arg %s', opt)
            sys.exit(3)
    if path is None:
        logger.critical('运行参数错误: 需要配置文件路径')
        sys.exit()
    else:
        try:
            load_config(path)
        except Exception as e:
            logger.critical('载入配置失败')
            logging.exception(e)
            sys.exit()
    sys.argv = [sys.argv[0]]
    logger.info('配置载入完成')
    app.run()
